@@ -44,6 +44,7 @@
 #include <sstream>
 #include <memory>
 #include <chrono>
+#include <map>
 
 #include "http-parser/http_parser.h"
 
@@ -219,7 +220,7 @@ std::string quote_string(const std::string &target);
 
 std::string format_hex(const unsigned char *s, size_t len);
 
-template <size_t N> std::string format_hex(const unsigned char (&s)[N]) {
+template <size_t N> std::string format_hex(const unsigned char(&s)[N]) {
   return format_hex(s, N);
 }
 
@@ -365,11 +366,11 @@ inline bool strieq(const char *a, const std::string &b) {
 }
 
 template <typename InputIt, size_t N>
-bool strieq_l(const char (&a)[N], InputIt b, size_t blen) {
+bool strieq_l(const char(&a)[N], InputIt b, size_t blen) {
   return strieq(a, N - 1, b, blen);
 }
 
-template <size_t N> bool strieq_l(const char (&a)[N], const std::string &b) {
+template <size_t N> bool strieq_l(const char(&a)[N], const std::string &b) {
   return strieq(a, N - 1, std::begin(b), b.size());
 }
 
@@ -399,10 +400,14 @@ inline bool streq(const char *a, const char *b) {
 }
 
 template <typename InputIt, size_t N>
-bool streq_l(const char (&a)[N], InputIt b, size_t blen) {
+bool streq_l(const char(&a)[N], InputIt b, size_t blen) {
   return streq(a, N - 1, b, blen);
 }
 
+template <size_t N> bool streq_l(const char(&a)[N], const std::string &b) {
+  return streq(a, N - 1, std::begin(b), b.size());
+}
+
 bool strifind(const char *a, const char *b);
 
 template <typename InputIt> void inp_strlower(InputIt first, InputIt last) {
@@ -706,6 +711,12 @@ uint32_t get_uint32(const uint8_t *data);
 // order and returns it in host byte order.
 uint64_t get_uint64(const uint8_t *data);
 
+// Reads mime types file (see /etc/mime.types), and stores extension
+// -> MIME type map in |res|.  This function returns 0 if it succeeds,
+// or -1.
+int read_mime_types(std::map<std::string, std::string> &res,
+                    const char *filename);
+
 } // namespace util
 
 } // namespace nghttp2
