@@ -59,7 +59,7 @@ void test_nghttp2_session_send_headers_push_reply(void);
 void test_nghttp2_session_send_rst_stream(void);
 void test_nghttp2_session_send_push_promise(void);
 void test_nghttp2_session_is_my_stream_id(void);
-void test_nghttp2_session_upgrade(void);
+void test_nghttp2_session_upgrade2(void);
 void test_nghttp2_session_reprioritize_stream(void);
 void test_nghttp2_session_reprioritize_stream_with_idle_stream_dep(void);
 void test_nghttp2_submit_data(void);
@@ -70,6 +70,7 @@ void test_nghttp2_submit_request_with_data(void);
 void test_nghttp2_submit_request_without_data(void);
 void test_nghttp2_submit_response_with_data(void);
 void test_nghttp2_submit_response_without_data(void);
+void test_nghttp2_submit_response_push_response(void);
 void test_nghttp2_submit_trailer(void);
 void test_nghttp2_submit_headers_start_stream(void);
 void test_nghttp2_submit_headers_reply(void);
@@ -116,6 +117,7 @@ void test_nghttp2_session_stream_attach_item(void);
 void test_nghttp2_session_stream_attach_item_subtree(void);
 void test_nghttp2_session_stream_get_state(void);
 void test_nghttp2_session_stream_get_something(void);
+void test_nghttp2_session_find_stream(void);
 void test_nghttp2_session_keep_closed_stream(void);
 void test_nghttp2_session_keep_idle_stream(void);
 void test_nghttp2_session_detach_idle_stream(void);
@@ -132,6 +134,7 @@ void test_nghttp2_session_on_begin_headers_temporal_failure(void);
 void test_nghttp2_session_defer_then_close(void);
 void test_nghttp2_session_detach_item_from_closed_stream(void);
 void test_nghttp2_session_flooding(void);
+void test_nghttp2_session_change_stream_priority(void);
 void test_nghttp2_http_mandatory_headers(void);
 void test_nghttp2_http_content_length(void);
 void test_nghttp2_http_content_length_mismatch(void);
@@ -141,5 +144,6 @@ void test_nghttp2_http_ignore_regular_header(void);
 void test_nghttp2_http_ignore_content_length(void);
 void test_nghttp2_http_record_request_method(void);
 void test_nghttp2_http_push_promise(void);
+void test_nghttp2_http_head_method_upgrade_workaround(void);
 
 #endif /* NGHTTP2_SESSION_TEST_H */
